#include <tadah/mlip/atom.h>
#include <tadah/mlip/analytics/analytics.h>
#include <tadah/mlip/analytics/statistics.h>
#include <cmath>    // for std::abs, std::sqrt, std::pow
#include <stdexcept>

namespace tadah {
namespace mlip {

Analytics::Analytics(const StructureDB &st, const StructureDB &stp)
  : st(st),
    stp(stp)
{
    if (st.size() != stp.size()) {
        throw std::runtime_error("Containers differ in size.");
    }
}

/**
 * @brief Compute per-file MAE of energy (in eV/atom).
 * Uses the new approach that sums over all structures in each file f.
 */
tadah::core::t_type Analytics::calc_e_mae() const
{
    // We return a tadah::core::t_type with length st.nFiles(), 
    // each element is the MAE for one file.
    tadah::core::t_type emae_vec(st.nFiles());

    for (size_t f = 0; f < st.nFiles(); ++f) 
    {
        double emae = 0.0;
        size_t N = 0; 
        size_t nStructs = st.nStructuresInFile(f);

        for (size_t i = 0; i < nStructs; ++i) {
            // st(f,i) and stp(f,i) must have the same #atoms 
            double e1 = st(f,i).energy;
            double e2 = stp(f,i).energy;
            size_t na = st(f,i).natoms();
            if (na == 0) continue; // avoid divide-by-zero

            emae += std::abs(e1 - e2) / na;
            N++;
        }
        // average over the structures in file f
        if (N > 0) emae /= N; 
        emae_vec(f) = emae;
    }
    return emae_vec;
}

/**
 * @brief Compute per-file MAE of forces (in eV/Å).
 */
tadah::core::t_type Analytics::calc_f_mae() const
{
    tadah::core::t_type fmae_vec(st.nFiles());
    for (size_t f = 0; f < st.nFiles(); ++f)
    {
        double fmae = 0.0;
        size_t N = 0; // total # of force components aggregated

        size_t nStructs = st.nStructuresInFile(f);
        for (size_t i = 0; i < nStructs; ++i) {
            const auto& s1 = st(f,i);
            const auto& s2 = stp(f,i);

            // sum over each atom’s 3 force components
            for (size_t a = 0; a < s1.natoms(); ++a) {
                double fx1 = s1(a).fx();
                double fy1 = s1(a).fy();
                double fz1 = s1(a).fz();
                double fx2 = s2(a).fx();
                double fy2 = s2(a).fy();
                double fz2 = s2(a).fz();

                fmae += std::abs(fx1 - fx2);
                fmae += std::abs(fy1 - fy2);
                fmae += std::abs(fz1 - fz2);
                N += 3;
            }
        }
        if (N > 0) fmae /= N;
        fmae_vec(f) = fmae;
    }
    return fmae_vec;
}

/**
 * @brief Compute per-file MAE of stress (in eV/Å³ or whichever units).
 *        We only collect the symmetric components (x,y) for x <= y.
 */
tadah::core::t_type Analytics::calc_s_mae() const
{
    tadah::core::t_type smae_vec(st.nFiles());
    for (size_t f = 0; f < st.nFiles(); ++f)
    {
        double smae = 0.0;
        size_t N = 0; // number of stress components collected

        size_t nStructs = st.nStructuresInFile(f);
        for (size_t i = 0; i < nStructs; ++i) {
            const auto& s1 = st(f,i);
            const auto& s2 = stp(f,i);

            for (size_t x = 0; x < 3; ++x) {
                for (size_t y = x; y < 3; ++y) {
                    double val1 = s1.stress(x,y);
                    double val2 = s2.stress(x,y);
                    smae += std::abs(val1 - val2);
                    N++;
                }
            }
        }
        if (N > 0) smae /= N;
        smae_vec(f) = smae;
    }
    return smae_vec;
}

/**
 * @brief Compute per-file relative RMSE of energy: 
 *        sqrt( 1/N * ∑ [ (E1 - E2)/E1 ]² ).
 */
tadah::core::t_type Analytics::calc_e_rrmse() const
{
    tadah::core::t_type errmse_vec(st.nFiles());
    for (size_t f = 0; f < st.nFiles(); ++f)
    {
        double sumSq = 0.0;  // sum of squares
        size_t N = 0; 
        size_t nStructs = st.nStructuresInFile(f);

        for (size_t i = 0; i < nStructs; ++i) {
            double Eref = st(f,i).energy;
            double Epred = stp(f,i).energy;

            // skip if Eref == 0 to avoid dividing by zero
            if (std::abs(Eref) > 1e-14) {
                double diffRel = (Eref - Epred)/Eref;
                sumSq += diffRel * diffRel;
                N++;
            }
        }
        double val = (N > 0) ? std::sqrt(sumSq / N) : 0.0;
        errmse_vec(f) = val;
    }
    return errmse_vec;
}

/**
 * @brief Compute per-file RMSE of energy (in eV/atom).
 */
tadah::core::t_type Analytics::calc_e_rmse() const
{
    tadah::core::t_type ermse_vec(st.nFiles());
    for (size_t f = 0; f < st.nFiles(); ++f)
    {
        double sumSq = 0.0;
        size_t N = 0;
        size_t nStructs = st.nStructuresInFile(f);

        for (size_t i = 0; i < nStructs; ++i) {
            double Eref = st(f,i).energy;
            double Epred = stp(f,i).energy;
            size_t na = st(f,i).natoms();
            if (na == 0) continue;

            double diff = (Eref - Epred) / na;
            sumSq += diff * diff;
            N++;
        }
        double val = (N > 0) ? std::sqrt(sumSq / N) : 0.0;
        ermse_vec(f) = val;
    }
    return ermse_vec;
}

/**
 * @brief Compute per-file RMSE of forces (in eV/Å).
 */
tadah::core::t_type Analytics::calc_f_rmse() const
{
    tadah::core::t_type frmse_vec(st.nFiles());
    for (size_t f = 0; f < st.nFiles(); ++f)
    {
        double sumSq = 0.0;
        size_t N = 0;

        size_t nStructs = st.nStructuresInFile(f);
        for (size_t i = 0; i < nStructs; ++i) {
            const auto & s1 = st(f,i);
            const auto & s2 = stp(f,i);

            for (size_t a = 0; a < s1.natoms(); ++a) {
                double fx1 = s1(a).fx(), fx2 = s2(a).fx();
                double fy1 = s1(a).fy(), fy2 = s2(a).fy();
                double fz1 = s1(a).fz(), fz2 = s2(a).fz();

                sumSq += (fx1 - fx2)*(fx1 - fx2);
                sumSq += (fy1 - fy2)*(fy1 - fy2);
                sumSq += (fz1 - fz2)*(fz1 - fz2);
                N += 3;
            }
        }
        if (N > 0) sumSq /= N;
        frmse_vec(f) = (N > 0) ? std::sqrt(sumSq) : 0.0;
    }
    return frmse_vec;
}

/**
 * @brief Compute per-file RMSE of stress (in eV/Å^3).
 */
tadah::core::t_type Analytics::calc_s_rmse() const
{
    tadah::core::t_type srmse_vec(st.nFiles());
    for (size_t f = 0; f < st.nFiles(); ++f)
    {
        double sumSq = 0.0;
        size_t N = 0;

        size_t nStructs = st.nStructuresInFile(f);
        for (size_t i = 0; i < nStructs; ++i) {
            const auto & s1 = st(f,i);
            const auto & s2 = stp(f,i);

            for (size_t x = 0; x < 3; ++x) {
                for (size_t y = x; y < 3; ++y) {
                    double val1 = s1.stress(x,y);
                    double val2 = s2.stress(x,y);
                    double diff = val1 - val2;
                    sumSq += diff * diff;
                    N++;
                }
            }
        }
        if (N > 0) sumSq /= N;
        srmse_vec(f) = (N > 0) ? std::sqrt(sumSq) : 0.0;
    }
    return srmse_vec;
}

/**
 * @brief Compute per-file R² of energy (comparing E/atom in each structure).
 */
tadah::core::t_type Analytics::calc_e_r_sq() const
{
    tadah::core::t_type e_r_sq_vec(st.nFiles());
    for (size_t f = 0; f < st.nFiles(); ++f)
    {
        size_t nStructs = st.nStructuresInFile(f);
        // Build vectors of observed and predicted (E/atom)
        tadah::core::t_type obs(nStructs), pred(nStructs);

        for (size_t i = 0; i < nStructs; ++i) {
            double Eref = st(f,i).energy;
            double Epred= stp(f,i).energy;
            size_t na = st(f,i).natoms();
            double denom = (na == 0 ? 1.0 : double(na));

            obs(i)  = Eref / denom;
            pred(i) = Epred / denom;
        }
        e_r_sq_vec(f) = Statistics::r_sq(obs, pred);
    }
    return e_r_sq_vec;
}

/**
 * @brief Compute per-file R² of forces (flattening all atoms’ fx,fy,fz).
 */
tadah::core::t_type Analytics::calc_f_r_sq() const
{
    tadah::core::t_type f_r_sq_vec(st.nFiles());
    for (size_t f = 0; f < st.nFiles(); ++f)
    {
        // First, compute total # of force components (3 * all atoms in file f)
        size_t totalAtoms = 0;
        size_t nStructs   = st.nStructuresInFile(f);
        for (size_t i = 0; i < nStructs; ++i) {
            totalAtoms += st(f,i).natoms();
        }
        size_t nForceComps = 3 * totalAtoms;

        tadah::core::t_type obs(nForceComps), pred(nForceComps);
        size_t idx = 0;

        // Fill obs/pred with fx, fy, fz
        for (size_t i = 0; i < nStructs; ++i) {
            const auto & s1 = st(f,i);
            const auto & s2 = stp(f,i);
            for (size_t a = 0; a < s1.natoms(); ++a) {
                obs(idx)   = s1(a).fx();
                pred(idx)  = s2(a).fx();
                idx++;

                obs(idx)   = s1(a).fy();
                pred(idx)  = s2(a).fy();
                idx++;

                obs(idx)   = s1(a).fz();
                pred(idx)  = s2(a).fz();
                idx++;
            }
        }
        f_r_sq_vec(f) = Statistics::r_sq(obs, pred);
    }
    return f_r_sq_vec;
}

/**
 * @brief Compute per-file R² of stress (collecting the 6 independent components).
 */
tadah::core::t_type Analytics::calc_s_r_sq() const
{
    tadah::core::t_type s_r_sq_vec(st.nFiles());
    for (size_t f = 0; f < st.nFiles(); ++f)
    {
        size_t nStructs = st.nStructuresInFile(f);
        // Each structure => 6 unique stress components (sxx, syy, szz, sxy, sxz, syz)
        tadah::core::t_type obs(6 * nStructs), pred(6 * nStructs);

        size_t idx = 0;
        for (size_t i = 0; i < nStructs; ++i) {
            const auto & s1 = st(f,i);
            const auto & s2 = stp(f,i);

            // Fill in (x,y) for x<=y
            for (size_t x = 0; x < 3; ++x) {
                for (size_t y = x; y < 3; ++y) {
                    obs(idx)  = s1.stress(x,y);
                    pred(idx) = s2.stress(x,y);
                    idx++;
                }
            }
        }
        s_r_sq_vec(f) = Statistics::r_sq(obs, pred);
    }
    return s_r_sq_vec;
}

} // namespace mlip
} // namespace tadah

