#include "tadah/mlip/design_matrix/functions/kernels/dm_kern_base.h"
#include "tadah/models/functions/function_base.h"
#include <tadah/mlip/design_matrix/functions/basis_functions/dm_bf_polynomial2.h>

namespace tadah {
namespace mlip {
DM_BF_Polynomial2::DM_BF_Polynomial2() {}
DM_BF_Polynomial2::DM_BF_Polynomial2(const tadah::core::Config &c): 
  tadah::models::Function_Base(c),
  DM_BF_Base(c),
  tadah::models::BF_Polynomial2(c)
{}
size_t DM_BF_Polynomial2::get_phi_cols(const tadah::core::Config &config)
{
  size_t cols = config.get<size_t>("DSIZE");
  return (cols*cols+cols)/2;
}
void DM_BF_Polynomial2::calc_phi_energy_row(tadah::core::phi_type &Phi,
                                            size_t &row,
                                            const double fac,
                                            const Structure &, const StructureNeighborView &st_nb,
                                            const StDescriptors &st_d)
{
  for (size_t a=0; a<st_d.naed();++a) {
    const tadah::core::aed_type& aed = st_d.get_aed(a);
    size_t b=0;
    for (size_t i=0; i<st_d.dim(); ++i) {
      for (size_t ii=i; ii<st_d.dim(); ++ii) {
        Phi(row,b++) += aed(i)*aed(ii)*fac;
      }
    }
  }
  row++;
}
void DM_BF_Polynomial2::calc_phi_force_rows(tadah::core::phi_type &Phi,
                                            size_t &row,
                                            const double fac,
                                            const Structure &st, const StructureNeighborView &st_nb,
                                            const StDescriptors &st_d)
{
  for (size_t a=0; a<st.natoms(); ++a) {

    const std::size_t* NbPtr = st_nb.getNeighborsPtr(a);
    size_t NNbrs = st_nb.numNeighbors(a);
    tadah::core::Vec3dSoAConstView ra = st.positionView(a);
    const tadah::core::aed_type& aedi = st_d.get_aed(a);
    for (size_t jj=0; jj<NNbrs; ++jj) {
      size_t gj = NbPtr[jj];
      size_t j= st.globalToLocal(gj);  // j is local to st
      size_t aa = st_nb.getAAIndex(a, j, jj);
      const tadah::core::fd_type &fdji = st_d.fd[j][aa];
      const tadah::core::fd_type &fdij = st_d.fd[a][jj];
      const tadah::core::aed_type& aedj = st_d.get_aed(j);
      for (size_t k=0; k<3; ++k) {
        size_t b=0;
        for (size_t i=0; i<fdij.rows(); ++i) {
          for (size_t ii=i; ii<fdij.rows(); ++ii) {
            Phi(row+k,b) -= fac*(fdij(i,k)*aedi(ii) + fdij(ii,k)*aedi(i)
              - fdji(i,k)*aedj(ii) - fdji(ii,k)*aedj(i));
            b++;
          }
        }
      }

    }
    row+=3;
  }
}
void DM_BF_Polynomial2::calc_phi_stress_rows(tadah::core::phi_type &Phi,
                                             size_t &row,
                                             const double fac[6],
                                             const Structure &st, const StructureNeighborView &st_nb, 
                                             const StDescriptors &st_d)
{
  double V_inv = 1 / st.get_volume();
  tadah::core::Vec3d rj;
  for (size_t a = 0; a < st.natoms(); ++a) {

    const std::size_t *NbPtr = st_nb.getNeighborsPtr(a);
    size_t NNbrs = st_nb.numNeighbors(a);
    tadah::core::Vec3dSoAConstView ra = st.positionView(a);
    const tadah::core::aed_type &aedi = st_d.get_aed(a);
    for (size_t jj = 0; jj < NNbrs; ++jj) {
      size_t gj = NbPtr[jj];
      size_t j = st.globalToLocal(gj); // j is local to st
      size_t aa = st_nb.getAAIndex(a, j, jj);

      auto sx = *st_nb.getShiftXPtr(j);
      auto sy = *st_nb.getShiftYPtr(j);
      auto sz = *st_nb.getShiftZPtr(j);

      rj(0) = st.x(j) + sx*st.cell(0,0) + sy*st.cell(0,1) + sz*st.cell(0,2);
      rj(1) = st.y(j) + sx*st.cell(1,0) + sy*st.cell(1,1) + sz*st.cell(1,2);
      rj(2) = st.z(j) + sx*st.cell(2,0) + sy*st.cell(2,1) + sz*st.cell(2,2);

      const tadah::core::fd_type &fdji = st_d.fd[j][aa];
      const tadah::core::fd_type &fdij = st_d.fd[a][jj];
      const tadah::core::aed_type &aedj = st_d.get_aed(j);
      size_t mn = 0;
      for (size_t x = 0; x < 3; ++x) {
        for (size_t y = x; y < 3; ++y) {
          size_t b = 0;
          for (size_t i = 0; i < fdij.rows(); ++i) {
            for (size_t ii = i; ii < fdij.rows(); ++ii) {
              Phi(row + mn, b) +=
                  V_inv * 0.5 * fac[mn] * (ra(x) - rj(x)) *
                  (fdij(i, y) * aedi(ii) + fdij(ii, y) * aedi(i) -
                   fdji(i, y) * aedj(ii) - fdji(ii, y) * aedj(i));
              b++;
            }
          }
          mn++;
        }
      }
    }
  }
  row += 6;
}

}
}
